#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
	vec2 uv2;
	mat3 TBN;
} vert;

layout (location = 0) out vec4 color;
layout (location = 1) out vec4 brightColor;

#define MAX_POINT_LIGHTS 35


uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;

uniform vec3 matColor;

uniform bool useTexture = true;
uniform sampler2D diffuseTexture;

uniform bool useSpecularMap = false;
uniform sampler2D specularMap;

uniform bool useNormalMap = false;
uniform sampler2D normalMap;

uniform bool useLightMap = false;
uniform sampler2D lightMap;

uniform bool useEmissionMap = false;
uniform sampler2D emissionMap;
uniform vec3 emissionColor;


uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
} pointL;

uniform struct PointLights{
	vec3 color[MAX_POINT_LIGHTS];
	vec3 position[MAX_POINT_LIGHTS];
	vec3 attenuation[MAX_POINT_LIGHTS];
	bool dynamic[MAX_POINT_LIGHTS];

	//for spotlights:
	vec3 dir[MAX_POINT_LIGHTS];
	float offCut[MAX_POINT_LIGHTS];
	float outerOffCut[MAX_POINT_LIGHTS];
} pointLights;
uniform unsigned int numOfPointLights;


vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	vec3 r = reflect(-l, n);
	return (diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att; 
}

/*
* Calculate the intensity factor for spotlights
* @param spotLightDir: spotlight's direction
* @param lightDir: vector pointing from frag position to light position (lightPos - fragPos)
* @param cutOff: inner circle angle in radians
* @param outerCutOff: outer circle angle in radians
*/
float calcSpotlightFactor(vec3 spotlightDir, vec3 lightDir, float cutOff, float outerCutOff){

	lightDir = normalize(lightDir);
	float theta = dot(lightDir, normalize(-spotlightDir));
	float epsilon = cutOff - outerCutOff;
	float intensity = clamp((theta - outerCutOff) / epsilon, 0.0, 1.0); 
	return intensity;
}



void main() {	
	vec3 n = normalize(vert.normal_world);
	vec3 v = normalize(camera_world - vert.position_world);

	if(useNormalMap) {
		n = texture(normalMap, vert.uv).rgb;
		n = normalize(vert.TBN * n);
	}

	float specF = useSpecularMap ? (texture(specularMap, vert.uv)).r : materialCoefficients.z;
	
	vec3 baseColor;
	if(useTexture)
		baseColor = texture(diffuseTexture, vert.uv).rgb;
	else
		baseColor = matColor;

	
	color = vec4(0,0,0,1);

	//apply lightmap if there is one
	if(useLightMap) {
		color.rgb += texture(lightMap, vert.uv2).rgb;
	}

	//apply ambient lighting if there is no lightmap
	if(!useLightMap) {
		color = vec4(baseColor * materialCoefficients.x, 1); // ambient	//todo: not if lightmap
	}

	
	// add directional light contribution	(if no lightmap)
	if(!useLightMap) {
		color.rgb += phong(n, -dirL.direction, v, dirL.color * baseColor, materialCoefficients.y, dirL.color, specF, specularAlpha, false, vec3(0));
	}

			
	// add point light contributions (if there is a lightmap we skip static lights)
	for (unsigned int i = 0; i < numOfPointLights; i++) {
		if(useLightMap && pointLights.dynamic[i] == false)
			continue;
		color.rgb += phong(n, pointLights.position[i] - vert.position_world, v, pointLights.color[i] * baseColor, materialCoefficients.y, pointLights.color[i], specF, specularAlpha, true, pointLights.attenuation[i]);
	}


	//add emission on top of everything
	color.rgb += useEmissionMap ? texture(emissionMap, vert.uv).rgb: emissionColor;
	

	float brightness = dot(color.rgb, vec3(0.2126, 0.7152, 0.0722));
	if(brightness > 0.8)
        brightColor = vec4(color.rgb, 1.0);
    else
        brightColor = vec4(0.0, 0.0, 0.0, 1.0);
}

